import pretty_midi

from fortepyan.viz import pianoroll as roll
from fortepyan.midi.structures import MidiFile, MidiPiece
from fortepyan.audio.render import midi_to_mp3, midi_to_wav

__all__ = ["roll", "MidiFile", "MidiPiece", "midi_to_wav", "midi_to_mp3"]
__version__ = "0.1.7"

# Pretty MIDI will throw an unwanted error for large files with high PPQ
# This is a workaround
# https://github.com/craffel/pretty-midi/issues/112
pretty_midi.pretty_midi.MAX_TICK = 1e10



from dataclasses import field, dataclass

import matplotlib
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pretty_midi import note_number_to_name

from fortepyan.midi.structures import MidiPiece

# These could be properties of the PianoRoll only
# not global configs
N_PITCHES = 128
RESOLUTION = 30


@dataclass
class PianoRoll:
    roll: np.array
    lowest_pitch: int
    highest_pitch: int
    duration: float
    max_value: int

    # Plot elements
    y_ticks: np.array = field(init=False)
    pitch_labels: list[str] = field(init=False)
    x_ticks: np.array = field(init=False)
    x_labels: list[str] = field(init=False)

    def __post_init__(self):
        # "Octave" mode for y-ticks
        self.y_ticks = np.arange(0, 128, 12, dtype=float)

        # Adding new line shifts the label up a little and positions
        # it nicely at the height where the note actually is
        self.pitch_labels = [f"{note_number_to_name(it)}\n" for it in self.y_ticks]

        # Move the ticks to land between the notes
        # (each note is 1-width and ticks by default are centered, ergo: 0.5 shift)
        self.y_ticks -= 0.5

        # Prepare x ticks and labels
        n_ticks = min(30, self.duration)
        step = np.ceil(self.duration / n_ticks) * RESOLUTION
        x_ticks = np.arange(0, step * n_ticks, step)
        self.x_ticks = np.round(x_ticks)
        self.x_labels = [round(xt / RESOLUTION) for xt in self.x_ticks]


def draw_pianoroll_with_velocities(midi_piece: MidiPiece, cmap: str = "GnBu"):
    fig, axes = plt.subplots(
        nrows=2,
        ncols=1,
        figsize=[16, 9],
        gridspec_kw={
            "height_ratios": [4, 1],
            "hspace": 0,
        },
    )
    piece = sanitize_midi_piece(midi_piece)
    draw_piano_roll(ax=axes[0], midi_piece=piece, cmap=cmap)
    draw_velocities(ax=axes[1], midi_piece=piece, cmap=cmap)

    sanitize_xticks(axes[1], piece)

    return fig


def prepare_piano_roll(
    df: pd.DataFrame,
    time_indicator: float = None,
    time_start: float = 0,
    time_end: float = None,
) -> PianoRoll:
    if not time_end:
        # We don't really need a full second roundup
        time_end = np.ceil(df.end.max())

    if time_end <= df.end.max():
        print("Warning, piano roll is not showing everythin!")

    # duration = time_end - time_start
    duration = time_end
    n_time_steps = RESOLUTION * int(duration)
    pianoroll = np.zeros((N_PITCHES, n_time_steps), np.uint8)

    # Adjust velocity color intensity to be sure it's visible
    min_value = 20
    max_value = 160

    for it, row in df.iterrows():
        note_on = row.start * RESOLUTION
        note_on = np.round(note_on).astype(int)

        note_end = row.end * RESOLUTION
        note_end = np.round(note_end).astype(int)
        pitch_idx = int(row.pitch)

        if time_indicator and note_on <= time_indicator * RESOLUTION < note_end:
            color_value = max_value
        else:
            color_value = min_value + row.velocity
        pianoroll[pitch_idx, note_on:note_end] = color_value

    # Could be a part of "prepare empty piano roll"
    for it in range(N_PITCHES):
        is_black = it % 12 in [1, 3, 6, 8, 10]
        if is_black:
            pianoroll[it, :] += min_value

    pianoroll = PianoRoll(
        roll=pianoroll,
        lowest_pitch=df.pitch.min(),
        highest_pitch=df.pitch.max(),
        duration=duration,
        max_value=max_value,
    )

    return pianoroll


def sanitize_midi_piece(piece: MidiPiece) -> MidiPiece:
    # 20 minutes?
    duration_threshold = 1200
    if piece.duration > duration_threshold:
        # TODO Logger
        print("Warning: playtime to long! Showing after trim")
        piece = piece.trim(0, duration_threshold)

    return piece


def sanitize_midi_frame(mf: pd.DataFrame) -> pd.DataFrame:
    # Do not modify input data
    df = mf.copy()

    # Make it start at 0.0
    df.end -= df.start.min()
    df.start -= df.start.min()
    duration_in = df.end.max()

    # 20 minutes?
    duration_threshold = 1200
    if duration_in > duration_threshold:
        # TODO Logger
        print("Warning: playtime to long! Showing after trim")
        ids = df.end <= duration_threshold
        df = df[ids]

    return df


def draw_piano_roll(
    ax: plt.Axes,
    midi_piece: MidiPiece,
    time: float = 0.0,
    cmap: str = "GnBu",
) -> plt.Axes:
    """
    Draws a pianoroll onto an ax.

    Parameters:
        ax: Matplotlib axis
        midi_piece: MidiPiece with piano performance
        time (Optional[float]): Use for dynamic visualization - will highlight notes
            played at this *time* value
        cmap (str): colormap recognizable by Matplotlib

    Returns:
        ax: Matplotlib axis with pianoroll.
    """
    piece = sanitize_midi_piece(midi_piece)
    piano_roll = prepare_piano_roll(piece.df, time_indicator=time)

    ax.imshow(
        piano_roll.roll,
        aspect="auto",
        vmin=0,
        vmax=138,
        origin="lower",
        interpolation="none",
        cmap=cmap,
    )

    ax.set_yticks(piano_roll.y_ticks)
    ax.set_yticklabels(piano_roll.pitch_labels, fontsize=15)

    # Show keyboard range where the music is
    y_min = piano_roll.lowest_pitch - 1
    y_max = piano_roll.highest_pitch + 1
    ax.set_ylim(y_min, y_max)

    ax.set_xticks(piano_roll.x_ticks)
    ax.set_xticklabels(piano_roll.x_labels, rotation=60)
    ax.set_xlabel("Time [s]")
    ax.set_xlim(0, piano_roll.duration * RESOLUTION)
    ax.grid()

    # Vertical position indicator
    ax.axvline(time * RESOLUTION, color="k", lw=0.5)

    return ax


def draw_velocities(ax: plt.Axes, midi_piece: MidiPiece, cmap: str = "GnBu") -> plt.Axes:
    piece = sanitize_midi_piece(midi_piece)
    df = piece.df
    colormap = matplotlib.cm.get_cmap(cmap)
    color = colormap(125 / 127)

    ax.plot(df.start, df.velocity, "o", ms=7, color=color)
    ax.plot(df.start, df.velocity, ".", color="white")
    ax.vlines(
        df.start,
        ymin=0,
        ymax=df.velocity,
        lw=2,
        alpha=0.777,
        colors=color,
    )
    ax.set_ylim(0, 128)

    return ax


def sanitize_xticks(ax: plt.Axes, piece: MidiPiece):
    """Sanitize the x-axis of a Matplotlib plot for easier readability.

    This function takes two parameters, `ax` and `piece`, which represent the Matplotlib axes object and the midi piece
    that the plot is based on, respectively. The function sets the x-axis tick positions, labels, and limits,
    and adds a grid to the plot to make it easier to read.

    Args:
    - ax (plt.Axes): Matplotlib axes object
    - piece (MidiPiece): `MidiPiece` object used to create the plot
    """
    # Calculate the number of seconds in the plot
    n_seconds = np.ceil(piece.duration)
    # Set the maximum number of x-axis ticks to 30
    n_ticks = min(30, n_seconds)
    # Calculate the step size for the x-axis tick positions
    step = np.ceil(n_seconds / n_ticks)
    # Calculate the x-axis tick positions
    x_ticks = np.arange(0, step * n_ticks, step)
    # Round the x-axis tick positions to the nearest integer
    x_ticks = np.round(x_ticks)
    # Set the x-axis tick labels to the same values as the tick positions
    labels = [xt for xt in x_ticks]

    # Set the x-axis tick positions and labels, and add a label to the x-axis
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(labels, rotation=60, fontsize=15)
    ax.set_xlabel("Time [s]")
    # Set the x-axis limits to the range of the data
    ax.set_xlim(0, n_seconds)
    # Add a grid to the plot
    ax.grid()



from dataclasses import field, dataclass

import numpy as np
import pretty_midi
import pandas as pd


@dataclass
class MidiPiece:
    df: pd.DataFrame
    sustain: pd.DataFrame
    source: dict

    def __rich_repr__(self):
        yield "MidiPiece"
        yield "notes", self.df.shape
        yield "sustain", self.sustain.shape
        yield "minutes", round(self.duration / 60, 2)

    @property
    def size(self) -> int:
        return self.df.shape[0]

    def trim(self, start: float, finish: float) -> "MidiPiece":
        """Trim the MidiPiece object between the specified start and finish time.

        This function takes two parameters, `start` and `finish`, which represent the start and end time in seconds,
        and returns a new MidiPiece object that contains only the notes within the specified time range.

        Args:
        - start (float): start time in seconds
        - finish (float): end time in seconds

        Returns:
        - MidiPiece: the trimmed MidiPiece object
        """
        # Filter the rows in the data frame that are within the specified start and end time
        ids = (self.df.start >= start) & (self.df.end <= finish)
        # Get the indices of the rows that meet the criteria
        idxs = np.where(ids)[0]
        # Get the start and end indices for the new MidiPiece object
        start = idxs[0]
        finish = idxs[-1] + 1
        # Slice the original MidiPiece object to create the trimmed MidiPiece object
        out = self[start:finish]
        # Return the trimmed MidiPiece object
        return out

    def __sanitize_get_index(self, index: slice) -> slice:
        if not isinstance(index, slice):
            raise TypeError("You can only get a part of MidiFile that has multiple notes: Index must be a slice")

        if not index.start:
            index = slice(0, index.stop)

        if not index.stop:
            index = slice(index.start, self.size)

        return index

    def __getitem__(self, index: slice) -> "MidiPiece":
        index = self.__sanitize_get_index(index)
        part = self.df[index].reset_index(drop=True)

        # +0.2 to make sure we get some sustain data at the end to ring out
        ids = (self.sustain.time >= part.start.min()) & (self.sustain.time <= part.end.max() + 0.2)
        sustain_part = self.sustain[ids].reset_index(drop=True)

        first_sound = part.start.min()
        sustain_part.time -= first_sound
        part.start -= first_sound
        part.end -= first_sound

        # Make sure the piece can always be track back to the original file exactly
        out_source = self.source
        out_source["start"] = self.source.get("start", 0) + index.start
        out_source["finish"] = self.source.get("finish", 0) + index.stop
        out_source["start_time"] = self.source.get("start_time", 0) + first_sound
        out = MidiPiece(df=part, sustain=sustain_part, source=out_source)

        return out

    @property
    def duration(self) -> float:
        duration = self.df.end.max() - self.df.start.min()
        return duration

    def to_midi(self, track_name: str = "piano"):
        track = pretty_midi.PrettyMIDI()
        piano = pretty_midi.Instrument(program=0, name=track_name)

        for it, row in self.df.iterrows():
            note = pretty_midi.Note(
                velocity=int(row.velocity),
                pitch=int(row.pitch),
                start=row.start,
                end=row.end,
            )
            piano.notes.append(note)

        cc = [pretty_midi.ControlChange(64, int(r.value), r.time) for _, r in self.sustain.iterrows()]
        piano.control_changes = cc

        track.instruments.append(piano)

        return track


@dataclass
class MidiFile:
    path: str
    df: pd.DataFrame = field(init=False)
    sustain: pd.DataFrame = field(init=False)
    _midi: pretty_midi.PrettyMIDI = field(init=False, repr=False)

    def __rich_repr__(self):
        yield "MidiFile"
        yield self.path
        yield "notes", self.df.shape
        yield "sustain", self.sustain.shape
        yield "minutes", round(self.duration / 60, 2)

    @property
    def duration(self) -> float:
        return self._midi.get_end_time()

    @property
    def notes(self):
        return self._midi.instruments[0].notes

    @property
    def control_changes(self):
        return self._midi.instruments[0].control_changes

    def __post_init__(self):
        self._midi = pretty_midi.PrettyMIDI(self.path)
        df = pd.DataFrame(
            {
                "pitch": [note.pitch for note in self.notes],
                "velocity": [note.velocity for note in self.notes],
                "start": [note.start for note in self.notes],
                "end": [note.end for note in self.notes],
            }
        )
        self.df = df.sort_values("start", ignore_index=True)

        # Sustain CC is 64
        sf = pd.DataFrame(
            {
                "time": [cc.time for cc in self.control_changes if cc.number == 64],
                "value": [cc.value for cc in self.control_changes if cc.number == 64],
            }
        )
        self.sustain = sf

    def __getitem__(self, index: slice) -> MidiPiece:
        return self.piece[index]
        if not isinstance(index, slice):
            raise TypeError("You can only get a part of MidiFile that has multiple notes: Index must be a slice")

        part = self.df[index].reset_index(drop=True)

        # +0.2 to make sure we get some sustain data at the end to ring out
        ids = (self.sustain.time >= part.start.min()) & (self.sustain.time <= part.end.max() + 0.2)
        sustain_part = self.sustain[ids].reset_index(drop=True)

        first_sound = part.start.min()
        sustain_part.time -= first_sound
        part.start -= first_sound
        part.end -= first_sound

        source = {
            "type": "MidiFile",
            "path": self.path,
        }
        out = MidiPiece(df=part, sustain=sustain_part, source=source)

        return out

    @property
    def piece(self) -> MidiPiece:
        source = {
            "type": "MidiFile",
            "path": self.path,
        }
        out = MidiPiece(
            df=self.df,
            sustain=self.sustain,
            source=source,
        )
        return out



import urllib
from pathlib import Path

import appdirs

from .. import config as C


def download_if_needed() -> Path:
    app_name = "fortepyan"
    app_author = "me"

    data_dir = appdirs.user_data_dir(app_name, app_author)
    data_dir = Path(data_dir)

    data_dir.mkdir(parents=True, exist_ok=True)

    file_path = data_dir / "soundfont.sf2"

    if not file_path.exists():
        print("Downloading SoundFont")
        urllib.request.urlretrieve(C.SOUNDFONT_URL, file_path)

    return file_path



import os

POSTGRES_DSN = os.environ.get("PSQL_DSN")
AWS_ENDPOINT = os.environ.get("AWS_ENDPOINT")

SOUNDFONT_URL = "https://github.com/Nospoko/soundfonts/raw/master/GeneralUserGSv1.471/GeneralUserGSv1.471.sf2"



import tempfile

import pretty_midi
from pydub import AudioSegment
from midi2audio import FluidSynth

from fortepyan.audio import soundfont


def midi_to_wav(midi: pretty_midi.PrettyMIDI, wavpath: str):
    # This will be deleted
    tmp_midi_path = tempfile.mkstemp(suffix=".mid")[1]

    # Add an silent event to make sure the final notes
    # have time to ring out
    end_time = midi.get_end_time() + 0.2
    pedal_off = pretty_midi.ControlChange(64, 0, end_time)
    midi.instruments[0].control_changes.append(pedal_off)

    midi.write(tmp_midi_path)

    sound_font_path = soundfont.download_if_needed()
    synth = FluidSynth(sound_font=sound_font_path)
    synth.midi_to_audio(tmp_midi_path, wavpath)


def midi_to_mp3(midi: pretty_midi.PrettyMIDI, mp3_path: str = None):
    # This will be deleted
    tmp_wav_path = tempfile.mkstemp(suffix=".wav")[1]
    midi_to_wav(midi=midi, wavpath=tmp_wav_path)

    # Wav to mp3
    if not mp3_path:
        mp3_path = tempfile.mkstemp(suffix=".mp3")[1]
    print("Rendering audio to file:", mp3_path)
    AudioSegment.from_wav(tmp_wav_path).export(mp3_path, format="mp3")

    return mp3_path



import tempfile
from pathlib import Path
import multiprocessing as mp

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from fortepyan.viz import pianoroll as roll
from fortepyan.midi.structures import MidiPiece


class PianoRollScene:
    def __init__(self, piece: MidiPiece, title: str, cmap: str = "GnBu"):
        self.axes = []
        self.content_dir = Path(tempfile.mkdtemp())

        self.frame_paths = []

        self.piece = piece
        self.title = title
        self.cmap = cmap

        f, axes = plt.subplots(
            nrows=2,
            ncols=1,
            figsize=[16, 9],
            gridspec_kw={
                "height_ratios": [4, 1],
                "hspace": 0,
            },
        )

        self.figure = f
        self.roll_ax = axes[0]
        self.velocity_ax = axes[1]
        self.axes = [self.roll_ax, self.velocity_ax]

    def draw_all_axes(self, time: float) -> None:
        self.draw_piano_roll(time)
        self.draw_velocities(time)
        roll.sanitize_xticks(self.velocity_ax, self.piece)

    def draw_piano_roll(self, time: float) -> None:
        roll.draw_piano_roll(
            ax=self.roll_ax,
            midi_piece=self.piece,
            cmap=self.cmap,
            time=time,
        )
        self.roll_ax.set_title(self.title, fontsize=20)

    def draw_velocities(self, time: float) -> None:
        roll.draw_velocities(
            ax=self.velocity_ax,
            midi_piece=self.piece,
            cmap=self.cmap,
        )

    def save_frame(self, savepath="tmp/tmp.png"):
        self.figure.tight_layout()

        self.figure.savefig(savepath)

        self.clean_figure()

    def clean_figure(self):
        for ax in self.axes:
            ax.clear()

    def animate_part(self, part: pd.DataFrame):
        for it, row in part.iterrows():
            time = row.time
            frame_counter = int(row.counter)
            self.draw(time)
            savepath = self.content_dir / f"{100000 + frame_counter}.png"
            self.save_frame(savepath)
            self.frame_paths.append(savepath)

    def draw(self, time: float) -> None:
        self.clean_figure()
        self.figure.tight_layout()
        self.draw_all_axes(time)

    def prepare_animation_steps(self, framerate: int = 30) -> pd.DataFrame:
        """
        Prepare the data required for the animation.

        Parameters:
            framerate (int): Framerate for the animation (default is 30).

        Returns:
            pd.DataFrame: DataFrame containing time and counter for each frame.
        """
        # Calculate the maximum time required for the animation
        max_time = np.ceil(self.piece.df.end.max()).astype(int)
        # Calculate the number of frames required for the animation
        n_frames = max_time * framerate
        # Create an array of times that will be used to create the animation
        times = np.linspace(0, max_time - 1 / framerate, n_frames)
        # Create a DataFrame to store the time and counter for each frame
        df = pd.DataFrame({"time": times, "counter": range(n_frames)})

        return df

    def render(self, framerate: int = 30) -> None:
        """
        Render the animation using a single process.

        Parameters:
            framerate (int): Framerate for the animation (default is 30).
        """
        df = self.prepare_animation_steps(framerate)

        # Call the animate_part function with the entire DataFrame as an argument
        self.animate_part(df)

        # Return the directory containing the generated animation content
        return self.content_dir

    def render_mp(self, framerate: int = 30) -> None:
        """
        Render the animation using multi-processing to speed up the process.

        Parameters:
            framerate (int): Framerate for the animation (default is 30).
        """
        df = self.prepare_animation_steps(framerate)

        # Step size for dividing the DataFrame into parts
        step = 50

        # Divide the DataFrame into parts based on the step size
        parts = [df[sta : sta + step] for sta in range(0, df.shape[0], step)]

        # Create a pool of processes using all available CPU cores
        with mp.Pool(mp.cpu_count()) as pool:
            # Map the animate_part function to each part of the DataFrame
            pool.map(self.animate_part, parts)

        # Return the directory containing the generated animation content
        return self.content_dir



import subprocess

from fortepyan.audio.render import midi_to_mp3
from fortepyan.midi.structures import MidiPiece
from fortepyan.animation import pianoroll as pianoroll_animation


def make_video(
    piece: MidiPiece,
    movie_path: str,
    title: str = "animation",
    cmap: str = "PuBuGn",
):
    scene = pianoroll_animation.PianoRollScene(piece, title=title, cmap=cmap)
    mp3_path = midi_to_mp3(piece.to_midi())

    scene_frames_dir = scene.render_mp()
    command = f"""
        ffmpeg -y -f image2 -framerate 30 -i {str(scene_frames_dir)}/10%4d.png\
        -loglevel quiet -i {mp3_path} -map 0:v:0 -map 1:a:0\
        {movie_path}
    """
    print("Rendering a movie to file:", movie_path)
    subprocess.call(command, shell=True)
